/* Coyright (c) 2004-2005  Petr Vandrovec.   Licensed under LGPL-2.1 */

#include "vmdsp_plugin.h"

#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <linux/soundcard.h>
#include <alsa/asoundlib.h>

struct alsa_plugin {
   struct vmdsp_plugin g;
   const char* server;
   const char* app;
};

struct alsa_dsp {
   struct vmdsp_dsp g;
   int speed;
   int stereo;
   int format;
   int mode;
   snd_pcm_t *rd;
   snd_pcm_t *wr;
};

#define container_of(ptr, type, member) ({                      \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})

#define INFO2ESD(x) container_of(x, struct alsa_plugin, g)
#define ESDP(x) struct alsa_plugin* alsa = INFO2ESD(x)

#define DSP2ESD(x) container_of(x, struct alsa_dsp, g)
#define ESDDSPP struct alsa_dsp* dspp = DSP2ESD(dsp)


#define USE_PERIOD_SIZE 1
#define INTERNAL(x) (x)
/**
 * \brief Set the hardware and software parameters in a simple way
 * \param pcm PCM handle
 * \param format required PCM format
 * \param access required PCM access
 * \param channels required PCM channels
 * \param rate required sample rate in Hz
 * \param soft_resample 0 = disallow alsa-lib resample stream, 1 = allow resampling
 * \param latency required overall latency in us (0 = optimum latency for players)
 * \return 0 on success otherwise a negative error code
 */
int my_snd_pcm_set_params(snd_pcm_t *pcm,
                       snd_pcm_format_t format,
                       snd_pcm_access_t access,
                       unsigned int channels,
                       unsigned int rate,
                       int soft_resample,
                       unsigned int latency)
{
        snd_pcm_hw_params_t *params;
        snd_pcm_sw_params_t *swparams;
        const char *s = snd_pcm_stream_name(snd_pcm_stream(pcm));
        snd_pcm_uframes_t buffer_size, period_size;
        unsigned int rrate, period_time;
        int err;

        snd_pcm_hw_params_alloca(&params);
        snd_pcm_sw_params_alloca(&swparams);

	assert(pcm);
	/* choose all parameters */
	err = snd_pcm_hw_params_any(pcm, params);
	if (err < 0) {
	        SNDERR("Broken configuration for %s: no configurations available", s);
	        return err;
        }
        /* set software resampling */
        err = snd_pcm_hw_params_set_rate_resample(pcm, params, soft_resample);
        if (err < 0) {
                SNDERR("Resampling setup failed for %s: %s", s, snd_strerror(err));
                return err;
        }
	/* set the selected read/write format */
	err = snd_pcm_hw_params_set_access(pcm, params, access);
	if (err < 0) {
		SNDERR("Access type not available for %s: %s", s, snd_strerror(err));
		return err;
	}
	/* set the sample format */
	err = snd_pcm_hw_params_set_format(pcm, params, format);
	if (err < 0) {
		SNDERR("Sample format not available for %s: %s", s, snd_strerror(err));
		return err;
	}
	/* set the count of channels */
	err = snd_pcm_hw_params_set_channels(pcm, params, channels);
	if (err < 0) {
		SNDERR("Channels count (%i) not available for %s: %s", s, channels, snd_strerror(err));
		return err;
	}
	/* set the stream rate */
	rrate = rate;
	err = INTERNAL(snd_pcm_hw_params_set_rate_near)(pcm, params, &rrate, 0);
	if (err < 0) {
		SNDERR("Rate %iHz not available for playback: %s", rate, snd_strerror(err));
		return err;
	}
	if (rrate != rate) {
		SNDERR("Rate doesn't match (requested %iHz, get %iHz)", rate, err);
		return -EINVAL;
	}

	/* set the buffer time */
	err = snd_pcm_hw_params_set_buffer_time_near(pcm, params, &latency, NULL);
	if (err < 0) {
	        /* error path -> set period size as first */
        	/* set the period time */
        	period_time = latency / 4;
        	err = INTERNAL(snd_pcm_hw_params_set_period_time_near)(pcm, params, &period_time, NULL);
        	if (err < 0) {
        		SNDERR("Unable to set period time %i for %s: %s", period_time, s, snd_strerror(err));
        		return err;
        	}
#if USE_PERIOD_SIZE
                err = INTERNAL(snd_pcm_hw_params_get_period_size)(params, &period_size, NULL);
                if (err < 0) {
                	SNDERR("Unable to get period size for %s: %s", s, snd_strerror(err));
                	return err;
        	}
        	buffer_size = period_size * 4;
        	err = INTERNAL(snd_pcm_hw_params_set_buffer_size_near)(pcm, params, &buffer_size);
                if (err < 0) {
                	SNDERR("Unable to set buffer size %lu %s: %s", buffer_size, s, snd_strerror(err));
                	return err;
        	}
        	err = INTERNAL(snd_pcm_hw_params_get_buffer_size)(params, &buffer_size);
        	if (err < 0) {
        		SNDERR("Unable to get buffer size for %s: %s", s, snd_strerror(err));
        		return err;
        	}
#endif
	} else {
	        /* standard configuration buffer_time -> periods */
        	err = snd_pcm_hw_params_get_buffer_size(params, &buffer_size);
        	if (err < 0) {
        		SNDERR("Unable to get buffer size for %s: %s", s, snd_strerror(err));
        		return err;
        	}
        	err = snd_pcm_hw_params_get_buffer_time(params, &latency, NULL);
        	if (err < 0) {
        		SNDERR("Unable to get buffer time (latency) for %s: %s", s, snd_strerror(err));
        		return err;
        	}
        	/* set the period time */
        	period_time = latency / 4;
        	err = snd_pcm_hw_params_set_period_time_near(pcm, params, &period_time, NULL);
        	if (err < 0) {
        		SNDERR("Unable to set period time %i for %s: %s", s, period_time, snd_strerror(err));
        		return err;
        	}
#if USE_PERIOD_SIZE
                err = snd_pcm_hw_params_get_period_size(params, &period_size, NULL);
                if (err < 0) {
                	SNDERR("Unable to get period size for %s: %s", s, snd_strerror(err));
                	return err;
        	}
#endif
        }
	/* write the parameters to device */
	err = snd_pcm_hw_params(pcm, params);
	if (err < 0) {
		SNDERR("Unable to set hw params for %s: %s", s, snd_strerror(err));
		return err;
	}

	/* get the current swparams */
	err = snd_pcm_sw_params_current(pcm, swparams);
	if (err < 0) {
		SNDERR("Unable to determine current swparams for %s: %s", s, snd_strerror(err));
		return err;
	}
	/* start the transfer when the buffer is almost full: */
	/* (buffer_size / avail_min) * avail_min */
#if USE_PERIOD_SIZE
	err = snd_pcm_sw_params_set_start_threshold(pcm, swparams, (buffer_size / period_size) * period_size);
	if (err < 0) {
		SNDERR("Unable to set start threshold mode for %s: %s", s, snd_strerror(err));
		return err;
	}
	/* allow the transfer when at least period_size samples can be processed */
	err = snd_pcm_sw_params_set_avail_min(pcm, swparams, period_size);
	if (err < 0) {
		SNDERR("Unable to set avail min for %s: %s", s, snd_strerror(err));
		return err;
	}
#endif
	/* align all transfers to 1 sample */
	err = snd_pcm_sw_params_set_xfer_align(pcm, swparams, 1);
	if (err < 0) {
		SNDERR("Unable to set transfer align for %s: %s", s, snd_strerror(err));
		return err;
	}
	/* write the parameters to the playback device */
	err = snd_pcm_sw_params(pcm, swparams);
	if (err < 0) {
		SNDERR("Unable to set sw params for %s: %s", s, snd_strerror(err));
		return err;
	}

	return 0;
}

static int
alsa_cleanup(struct vmdsp_plugin* info) {
   ESDP(info);

   free(alsa);
   return 0;
}

static struct vmdsp_dsp*
alsa_new_dsp(struct vmdsp_plugin* info, const char* name, int mode) {
   ESDP(info);
   struct alsa_dsp* dspp;
   
   dspp = malloc(sizeof(*dspp));
   if (dspp) {
      int fd;

      memset(dspp, 0, sizeof(*dspp));
      dspp->g.info = &alsa->g;

      fd = vmdsp_open64("/dev/null", mode, 0);
      if (fd < 0) { //open error
         free(dspp);
         errno = ENODEV;
         return NULL;
      }
      dspp->g.fd = fd;

      dspp->speed = 22050;
      dspp->stereo = 0;
      dspp->format = AFMT_U8;

      dspp->mode = mode;
      return &dspp->g;
   }
   return NULL;
   (void)name;
}

static int
alsa_close_dsp(struct vmdsp_dsp* dsp) {
   ESDDSPP;

   if (dspp->rd) {
      snd_pcm_close(dspp->rd);
   }
   if (dspp->wr) {
      snd_pcm_close(dspp->wr);
      printf("output pcm closed\n");
   }

   vmdsp_close(dspp->g.fd);
   free(dspp);
   return 0;
}

static int
dspp_start_playback(struct alsa_dsp* dspp) {
   ESDP(dspp->g.info);

   if (dspp->mode != O_RDWR && dspp->mode != O_WRONLY) {
      vmdsp_log("Would not start playback: bad mode\n");
      errno = EINVAL;
      return -1;
   }

   snd_pcm_open(&(dspp->wr), "default", SND_PCM_STREAM_PLAYBACK, 0); /* blocking */

   printf("format: %d stereo: %d speed %d\n", dspp->format, dspp->stereo, dspp->speed);
   my_snd_pcm_set_params(dspp->wr, dspp->format, SND_PCM_ACCESS_RW_INTERLEAVED,
                      dspp->stereo?2:1, dspp->speed, 1, 5000000);

   if (!dspp->wr) {
      vmdsp_log("Would not start playback: stream open failed\n");
      errno = EINVAL;
      return -1;
   }
   vmdsp_log("Playback stream %p opened\n", dspp->wr);
   return 0;
}

static int
dspp_start_record(struct alsa_dsp* dspp) {
   errno = EINVAL;
   return -1;
   (void)dspp;
}

static ssize_t
dspp_read(struct vmdsp_dsp* dsp, void* buffer, size_t length) {
   ESDDSPP;

   if (!dspp->rd) {
      if (dspp_start_record(dspp)) {
	 return -1;
      }
   }
   if (length > 512) {
      length = 512;
   }
   
   return snd_pcm_readi(dspp->rd, buffer, length); //xxx fix me
}

static ssize_t
dspp_write(struct vmdsp_dsp* dsp, const void* buffer, size_t length) {
   ESDDSPP;
   ssize_t alsa_written = 0;

   if (!dspp->wr) {
      if (dspp_start_playback(dspp)) {
	 return length;
      }
   }


   while(length > 0){
	   int frames;
	   snd_pcm_sframes_t written_frames;

	   frames = snd_pcm_bytes_to_frames(dspp->wr, length);
	   written_frames = snd_pcm_writei(dspp->wr, buffer, frames);

	   if(written_frames > 0){
		   int written = snd_pcm_frames_to_bytes(dspp->wr, written_frames);
		   length -= written;
		   buffer += written;
		   alsa_written += written;
	   }
	   else if (written_frames <= 0) {
		   snd_pcm_prepare(dspp->wr);
//		   break;
	   }
   }

   return alsa_written;
}

static int
dspp_set_speed(struct vmdsp_dsp* dsp, int speed) {
   ESDDSPP;

   if (dspp->speed == speed) {
      return 0;
   }
   if (dspp->rd || dspp->wr) {
      vmdsp_warn("Changing sample speed while running (%u => %u)\n", dspp->speed, speed);
   }
   dspp->speed = speed;
   return 0;
}

static int
dspp_set_stereo(struct vmdsp_dsp* dsp, int stereo) {
   ESDDSPP;

   if (dspp->stereo == stereo) {
      return 0;
   }
   if (dspp->rd || dspp->wr) {
      vmdsp_warn("Changing stereo while running (%u => %u)\n", dspp->stereo, stereo);
   }
   dspp->stereo = stereo;
   return 0;
}

static int
dspp_set_format(struct vmdsp_dsp* dsp, int format) {
   ESDDSPP;

   switch (format) {
      case AFMT_U8:
          format = SND_PCM_FORMAT_U8;
      case AFMT_S16_LE:
          format = SND_PCM_FORMAT_S16_LE;
         break;
      default:
         format = SND_PCM_FORMAT_S16_LE;
	 break;
   }

   if (dspp->format == format) {
      return 0;
   }
   if (dspp->rd || dspp->wr) {
      vmdsp_warn("Changing format while running (%08X => %08X)\n", dspp->format, format);
   }
   dspp->format = format;
   return 0;
}

static int
dspp_get_format(struct vmdsp_dsp* dsp, int* format) {
   ESDDSPP;

   *format = dspp->format;
   return 0;
}

static const struct vmdsp_operations esd_ops = {
   .cleanup	= alsa_cleanup,
   .new_dsp	= alsa_new_dsp,
   .new_mixer	= NULL,
   .close_dsp	= alsa_close_dsp,
   .close_mixer	= NULL,
   .read	= dspp_read,
   .write	= dspp_write,
   .set_speed	= dspp_set_speed,
   .set_stereo	= dspp_set_stereo,
   .set_format	= dspp_set_format,
   .get_format  = dspp_get_format,
   .set_volume	= NULL,
   .get_volume	= NULL,
   .set_recsrc	= NULL,
   .get_recsrc	= NULL,
};

struct vmdsp_plugin*
vmdsp_plugin_init(const char* cfgfile) {
   struct alsa_plugin* esdp;

   esdp = malloc(sizeof(*esdp));
   if (!esdp) {
      return NULL;
   }
   memset(esdp, 0, sizeof(*esdp));
   esdp->g.d_op = &esd_ops;
   esdp->app = cfgfile;
   return &esdp->g;
}
